#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <math.h>
#include <mpi.h>
#include <hdf5.h>

#include "rundata.h"
#include "options.h"


void writehdf5file(rundata_t rundata, float *x, double *data) {
    /* identifiers */
    hid_t file_id, arr_group_id, data_dataset_id, x_dataset_id;
    hid_t data_dataspace_id, x_dataspace_id;
    hid_t loc_data_dataspace_id, loc_x_dataspace_id;
    hid_t globaldataspace,globalxspace;
    hid_t dist_id;
    hid_t fap_id;

    /* sizes */
    hsize_t datadims[1], xdims[1];
    hsize_t locdatadims[1], locxdims[1];

    /* MPI-IO hints for performance */
    MPI_Info info;

    /* parameters of the hyperslab */
    hsize_t counts[1];
    hsize_t strides[1];
    hsize_t offsets[1];
    hsize_t blocks[1];

    /* set the MPI-IO hints for better performance on GPFS */
    MPI_Info_create(&info);
    MPI_Info_set(info,"IBM_largeblock_io","true");

    /* Set up the parallel environment for file access*/
    fap_id = H5Pcreate(H5P_FILE_ACCESS);
    /* Include the file access property with IBM hint */
    H5Pset_fapl_mpio(fap_id, MPI_COMM_WORLD, info);

    /* Set up the parallel environment */
    dist_id = H5Pcreate(H5P_DATASET_XFER);
    /* we'll be writing collectixy */
    H5Pset_dxpl_mpio(dist_id, H5FD_MPIO_COLLECTIVE);

    /* Create a new file - truncate anything existing, use default properties */
    file_id = H5Fcreate(rundata.filename, H5F_ACC_TRUNC, H5P_DEFAULT, fap_id);

    /* HDF5 routines generally return a negative number on failure.  
     * Should check return values! */
    if (file_id < 0) {
        fprintf(stderr,"Could not open file %s\n", rundata.filename);
        return;
    }

    /* Create a new group within the new file */
    arr_group_id = H5Gcreate(file_id,"/ArrayData", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

    /* Create the data space for the two global datasets. */
    datadims[0] = rundata.globaln; 
    xdims[0]    = rundata.globaln; 
    
    data_dataspace_id = H5Screate_simple(1, datadims, NULL);
    x_dataspace_id  = H5Screate_simple(1, xdims,  NULL);

    /* Create the datasets within the file. 
     * H5T_IEEE_F64LE is a standard (IEEE) double precision (64 bit) floating (F) data type
     * and will work on any machine.  H5T_NATIVE_DOUBLE would work too, but would give
     * different results on GPC and TCS */

    data_dataset_id = H5Dcreate(file_id, "/ArrayData/Data", H5T_IEEE_F64LE, 
                                data_dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
    x_dataset_id  = H5Dcreate(file_id, "/ArrayData/x Coordinates",  H5T_IEEE_F32LE, 
                                x_dataspace_id,  H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

    /* Now create the data space for our sub-regions.   These are the data spaces
     * of our actual local data in memory. */
    locdatadims[0] = rundata.localn;
    locxdims[0] =    rundata.localn;
    
    loc_data_dataspace_id = H5Screate_simple(1, locdatadims, NULL);
    loc_x_dataspace_id    = H5Screate_simple(1, locxdims,  NULL);

     /* regions of teh global space are described in terms of "hyperslabs"
      * MPI_Type_create_vector()'s:  offset/block/stride/count    */

    offsets[0] = rundata.start;
    counts[0]  = rundata.localn;
    blocks[0]  = 1;
    strides[0] = 1;

    /* select this subset of the dataity variable's space in the file */
    globaldataspace = H5Dget_space(data_dataset_id);
    H5Sselect_hyperslab(globaldataspace, H5S_SELECT_SET, offsets, strides, counts, blocks);

    globalxspace = H5Dget_space(x_dataset_id);
    H5Sselect_hyperslab(globalxspace, H5S_SELECT_SET, offsets, strides, counts, blocks);

    /* Write the data.  We're writing it from memory, where it is saved 
     * in NATIVE_DOUBLE format */
    H5Dwrite(data_dataset_id, H5T_NATIVE_DOUBLE, loc_data_dataspace_id, globaldataspace, dist_id, data);
    H5Dwrite(x_dataset_id,    H5T_NATIVE_FLOAT,  loc_x_dataspace_id,    globalxspace,    dist_id, x);

    /* We'll create another group for related info and put some things in there */

    {
        hid_t other_group_id;
        hid_t timestep_id, timestep_space;
        hid_t comptime_id, comptime_space;
        int timestep=13;
        float comptime=81.773;

        /* create group */
        other_group_id = H5Gcreate(file_id,"/OtherStuff", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

        /* scalar space, data for integer timestep */
        timestep_space = H5Screate(H5S_SCALAR);
        timestep_id = H5Dcreate(other_group_id, "Timestep", H5T_STD_U32LE, 
                                timestep_space, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
        H5Dwrite(timestep_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &timestep);
        H5Dclose(timestep_id);
        H5Sclose(timestep_space);
   
        /* scalar space, data for floating compute time */
        comptime_space = H5Screate(H5S_SCALAR);
        comptime_id = H5Dcreate(other_group_id, "Compute Time", H5T_IEEE_F32LE, 
                                comptime_space, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
        H5Dwrite(comptime_id, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &comptime);
        H5Dclose(comptime_id);
        H5Sclose(comptime_space);
   
        H5Gclose(other_group_id);
    } 


    /* End access to groups & data sets and release resources used by them */
    H5Sclose(data_dataspace_id);
    H5Dclose(data_dataset_id);
    H5Sclose(x_dataspace_id);
    H5Dclose(x_dataset_id);
    H5Gclose(arr_group_id);
    H5Pclose(fap_id);
    H5Pclose(dist_id);
    
    /* Close the file */
    H5Fclose(file_id);
    return;
}


int main(int argc, char **argv) {
    int ierr, rank, size;
    float *x;
    double *data;
    rundata_t rundata;

    rundata.globaln = 100;
    strcpy(rundata.filename,"sine-hdf5.h5");

    ierr = MPI_Init(&argc, &argv);
    ierr|= MPI_Comm_size(MPI_COMM_WORLD, &size);
    ierr|= MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
    rundata.nprocs = size;
    rundata.rank = rank;

    get_options(argc,argv,&rundata);

    printf("[%d]: Assigned to region (%d,%d) \n", rank, rundata.start, rundata.start+rundata.localn-1);

    x = malloc(rundata.localn * sizeof(float));
    for (int i=0; i<rundata.localn; i++) 
        x[i] = (i+rundata.start)*rundata.dx;
    
    data = malloc(rundata.localn * sizeof(double));
    for (int i=0; i<rundata.localn; i++)
        data[i] = sin(x[i]);

    writehdf5file(rundata, x, data);

    free(data);
    free(x);
    MPI_Finalize();
    return 0;
}
