#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <math.h>
#include <time.h>
#include <mpi.h>
#include <netcdf.h>
#include <netcdf_par.h>

#include "rundata.h"
#include "options.h"

void writenetcdffile(rundata_t rundata, double *data) {
    /* identifiers */
    int file_id, xdim_id;
    int data_id;
    

    /* sizes */
    int datadims[1];

    /* name of unit for data */
    const char *dataunit="m";

    /* offsets for sub-regions of arrays */
    size_t starts[1];
    size_t counts[1];

    /* return status */
    int status;

    /* MPI-IO hints for performance */
    MPI_Info info = MPI_INFO_NULL;

    /* Create a new file - clobber anything existing */
    MPI_Comm comm = MPI_COMM_WORLD;
    status = nc_create_par(rundata.filename, NC_MPIIO | NC_NETCDF4 | NC_CLOBBER, comm, 
                               info, &file_id);

    /* netCDF routines return NC_NOERR on success */
    if (status != NC_NOERR) {
        fprintf(stderr,"Could not open file %s\n", rundata.filename);
        return;
    }

    /* define the dimensions */
    nc_def_dim(file_id, "X", rundata.globaln, &xdim_id);
    datadims[0] = xdim_id;

    /* now that we've defined the dimensions, we can define variables on them */
    nc_def_var(file_id, "Data",  NC_DOUBLE, 1, datadims, &data_id);

    /* assign units to the variables */
    nc_put_att_text(file_id, data_id, "units", strlen(dataunit), dataunit);

    /* we are now done defining variables and their attributes */
    nc_enddef(file_id);

    /* The big data will be written to collectively; eg write_all
     * the alternative is NC_INDEPENDENT  */
    nc_var_par_access(file_id, data_id,    NC_COLLECTIVE);

    starts[0] = (rundata.start);
    counts[0]  = rundata.localn;

    nc_put_vara_double(file_id, data_id,   starts, counts, &(data[0]));

    nc_close(file_id);
    return;
}


int get_options(int argc, char **argv, rundata_t *rundata);

int main(int argc, char **argv) {
    double *data;
    rundata_t rundata;
    int rank, size;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD,&size);
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);

    /*  
     * set default values for parameters, then check what the user says
     * with the options 
     */
    
    rundata.globaln = 100;
    rundata.nprocs = size;
    rundata.rank = rank;
    strcpy(rundata.filename,"sine.nc");

    /* get options */
    get_options(argc,argv,&rundata);

    printf("[%d]: Assigned to region (%d,%d) \n", rank, rundata.start, rundata.start+rundata.localn-1);
    
    /* set up variable */
    data = (double *)malloc(rundata.localn * sizeof(double));
    for (int i=0; i<rundata.localn; i++) 
        data[i] = sin( (i+rundata.start)*rundata.dx );
    
    writenetcdffile(rundata, data);
    printf("[%d]: Wrote file\n", rank);

    free(data);

    MPI_Finalize();
    return 0;
}  
